/* -*- mode: C; mode: folding; fill-column: 70; -*- */
/* Copyright 2010,  Georgia Institute of Technology, USA. */
/* See COPYING for license. */
#define _FILE_OFFSET_BITS 64
#define _THREAD_SAFE
#include "../compat.h"
#include <stdlib.h>
#include <stdio.h>
#include <upc_relaxed.h>
#include <string.h>
#include <math.h>

#include <assert.h>

static int64_t int64_casval(int64_t* p, int64_t oldval, int64_t newval);

#include "../graph500.h"
#include "../xalloc.h"
#include "../generator/graph_generator.h"

#define MINVECT_SIZE 2

static shared int64_t maxvtx, nv, sz;
static shared int64_t * restrict xoff; /* Length 2*nv+2 */
static shared int64_t * restrict xadjstore; /* Length MINVECT_SIZE + (xoff[nv] == nedge) */
static shared int64_t * restrict xadj;

shared int err;
shared int64_t *buf;

upc_lock_t *lock1, *lock2, *lock3, *lock4;

static void
find_nv (const struct packed_edge * restrict IJ, const int64_t nedge)
{

  lock1 = upc_all_lock_alloc();

  if(MYTHREAD==0)
  	maxvtx = -1;

	upc_barrier;

    int64_t k, gmaxvtx, tmaxvtx = -1;

      upc_forall (k = 0; k < nedge; ++k; k) {
	if (get_v0_from_edge(&IJ[k]) > tmaxvtx)
	  tmaxvtx = get_v0_from_edge(&IJ[k]);
	if (get_v1_from_edge(&IJ[k]) > tmaxvtx)
	  tmaxvtx = get_v1_from_edge(&IJ[k]);
      }

	    upc_lock(lock1);
	    gmaxvtx = maxvtx;
	    while (tmaxvtx > gmaxvtx)
	      gmaxvtx = int64_casval ((int64_t *)&maxvtx, gmaxvtx, tmaxvtx);
	    upc_unlock(lock1);

	upc_barrier;

	if(MYTHREAD==0){

	    nv = 1+maxvtx;
    }

    upc_barrier;
}

static int
alloc_graph (int64_t nedge)
{
  sz = (2*nv+2) * sizeof (*xoff);
  xoff = (shared int64_t *)upc_all_alloc(1, sz); //xmalloc_large_ext (sz);
  if (!xoff) return -1;
  return 0;
}

static void
free_graph (void)
{
  xfree_large ((int64_t *)&xadjstore);
  xfree_large ((int64_t *)&xoff);
}

#define XOFF(k) (xoff[2*(k)])
#define XENDOFF(k) (xoff[1+2*(k)])

static int64_t
prefix_sum (shared int64_t *buff)
{
  int nt, tid;
  int64_t slice_begin, slice_end, t1, t2, k;

  t1 = nv / THREADS;
  t2 = nv % THREADS;
  slice_begin = t1 * MYTHREAD + (MYTHREAD < t2? MYTHREAD : t2);
  slice_end = t1 * (MYTHREAD+1) + ((MYTHREAD+1) < t2? (MYTHREAD+1) : t2);

  buff[MYTHREAD] = 0;
  for (k = slice_begin; k < slice_end; ++k)
    buff[MYTHREAD] += XOFF(k);

  //upc_barrier;
  if(MYTHREAD==0){
    for (k = 1; k < THREADS; ++k)
      buff[k] += buff[k-1];
  }
  if (MYTHREAD)
    t1 = buff[MYTHREAD-1];
  else
    t1 = 0;

  for (k = slice_begin; k < slice_end; ++k) {
    int64_t tmp = XOFF(k);
    XOFF(k) = t1;
    t1 += tmp;
  }
  //upc_barrier;

  return buff[THREADS-1];
}

static int
setup_deg_off (const struct packed_edge * restrict IJ, int64_t nedge)
{
  int64_t k, accum;

  lock2 = upc_all_lock_alloc();

  lock3 = upc_all_lock_alloc();

  lock4 = upc_all_lock_alloc();

  upc_forall (k = 0; k < 2*nv+2; ++k; k)
    xoff[k] = 0;
  upc_forall (k = 0; k < nedge; ++k; k) {
    int64_t i = get_v0_from_edge(&IJ[k]);
    int64_t j = get_v1_from_edge(&IJ[k]);
    if (i != j) { /* Skip self-edges. */
      if (i >= 0){
	upc_lock(lock2);
	++XOFF(i);
	upc_unlock(lock2);
      }
      if (j >= 0){
	upc_lock(lock3);
	++XOFF(j);
	upc_unlock(lock3);
      }
    }
  }

  upc_barrier;
  //if(MYTHREAD==0) {
      buf = /*alloca (THREADS * sizeof (*buf));*/ (shared int64_t *)upc_all_alloc(THREADS, sizeof (*buf));
      if (!buf) {
	perror ("alloca for prefix-sum hosed");
	abort ();
      }
    //}

  upc_barrier;

  upc_forall (k = 0; k < nv; ++k; k)
	if (XOFF(k) < MINVECT_SIZE) XOFF(k) = MINVECT_SIZE;

  upc_barrier;

  upc_lock(lock4);
  accum = prefix_sum ((shared int64_t *)buf);
  upc_unlock(lock4);

  upc_forall (k = 0; k < nv; ++k; k)
	XENDOFF(k) = XOFF(k);

  if(MYTHREAD==0) {
      XOFF(nv) = accum;  printf("\n%ld\n", accum);
    }

  upc_barrier;

  if (!(xadjstore = (shared int64_t *)upc_all_alloc(1, (XOFF(nv) + MINVECT_SIZE) * sizeof (*xadjstore)))) //xmalloc_large_ext ((XOFF(nv) + MINVECT_SIZE) * sizeof (*xadjstore))))
	err = -1;

  upc_barrier;

  if (!err) {
	if(MYTHREAD==0) {
	xadj = &xadjstore[MINVECT_SIZE]; /* Cheat and permit xadj[-1] to work. */ 
	for (k = 0; k < XOFF(nv) + MINVECT_SIZE; ++k)
	  xadjstore[k] = -1;
	}
      }


 upc_barrier;

  if(MYTHREAD==0){
	printf("\n%ld\n", !xadj);
	printf("\n%ld\n", xadj);
	printf("\n%ld\n", !0);
	printf("\n%ld\n", xadjstore);
  	return !xadj;
	}
}

static void
scatter_edge (const int64_t i, const int64_t j)
{
  int64_t where;
  where = XENDOFF(i)++;
  xadj[where] = j;
}

static int
i64cmp (const void *a, const void *b)
{
  const int64_t ia = *(const int64_t*)a;
  const int64_t ib = *(const int64_t*)b;
  if (ia < ib) return -1;
  if (ia > ib) return 1;
  return 0;
}

static void
pack_vtx_edges (const int64_t i)
{
  int64_t kcur, k;
  if (XOFF(i)+1 >= XENDOFF(i)) return;
  qsort ((int64_t*)&xadj[XOFF(i)], XENDOFF(i)-XOFF(i), sizeof(*xadj), i64cmp);
  kcur = XOFF(i);
  for (k = XOFF(i)+1; k < XENDOFF(i); ++k)
    if (xadj[k] != xadj[kcur])
      xadj[++kcur] = xadj[k];
  ++kcur;
  for (k = kcur; k < XENDOFF(i); ++k)
    xadj[k] = -1;
  XENDOFF(i) = kcur;
}

static void
pack_edges (void)
{
  int64_t v;

  upc_forall (v = 0; v < nv; ++v; v)
    pack_vtx_edges (v);
}

static void
gather_edges (const struct packed_edge * restrict IJ, int64_t nedge)
{
  int64_t k;

  for (k = 0; k < nedge; ++k) {
    int64_t i = get_v0_from_edge(&IJ[k]);
    int64_t j = get_v1_from_edge(&IJ[k]);
    if (i >= 0 && j >= 0 && i != j) {
      scatter_edge (i, j);
      scatter_edge (j, i);
    }
  }

  pack_edges ();
}

int 
create_graph_from_edgelist (struct packed_edge *IJ, int64_t nedge)
{
  find_nv (IJ, nedge);
  if (alloc_graph (nedge)) return -1;
  if (setup_deg_off (IJ, nedge)) {
    xfree_large ((int64_t *)&xoff);
    return -1;
  }
  gather_edges (IJ, nedge);
  return 0;
}

int
make_bfs_tree (int64_t *bfs_tree_out, int64_t *max_vtx_out,
	       int64_t srcvtx)
{
  int64_t * restrict bfs_tree = bfs_tree_out;
  int err = 0;

  int64_t * restrict vlist = NULL;
  int64_t k1, k2;

  *max_vtx_out = maxvtx;

  vlist = xmalloc_large (nv * sizeof (*vlist));
  if (!vlist) return -1;

  for (k1 = 0; k1 < nv; ++k1)
    bfs_tree[k1] = -1;

  vlist[0] = srcvtx;
  bfs_tree[srcvtx] = srcvtx;
  k1 = 0; k2 = 1;
  while (k1 != k2) {
    const int64_t oldk2 = k2;
    int64_t k;
    for (k = k1; k < oldk2; ++k) {
      const int64_t v = vlist[k];
      const int64_t veo = XENDOFF(v);
      int64_t vo;
      for (vo = XOFF(v); vo < veo; ++vo) {
	const int64_t j = xadj[vo];
	if (bfs_tree[j] == -1) {
	  bfs_tree[j] = v;
	  vlist[k2++] = j;
	}
      }
    }
    k1 = oldk2;
  }

  xfree_large (vlist);

  return err;
}

void
destroy_graph (void)
{
  free_graph ();
}

int64_t
int64_casval(int64_t* p, int64_t oldval, int64_t newval)
{
  int64_t v = *p;

  if (v == oldval)
    *p = newval;

  return v;
}

